# Licensed under GPL version 3 - see LICENSE.rst
''' Lynx telescope and instrument definition.

Since some aspects of the telescope and not well-defined yet, there are some rough edges.

Below is the running list of input data that could be improved with no or little code changes:

- QE (and entrance filters if any are present) for zero-order instruments missing
- Filter curves and QE for XGS detectors only defined to 2.1 keV. Just added arbitrary numbers
  after that to avoid overflow errors in the interpolation.
- Check mirror model. Does is use the energy dependence for different shells?
- Update L1 diffraction efficiencies. Note that transmission through L1 is currently handled separately.
- Update L2 dimensions.
'''

import copy
from astropy.table import Table
import astropy.units as u
from astropy.utils.data import get_pkg_data_filename
import numpy as np
import transforms3d
from scipy.interpolate import interp1d

from marxs import optics, simulator, design, analysis
from marxs.design.rowland import (design_tilted_torus, RowlandTorus,
                                  CircularMeshGrid)
from marxs.math.geometry import Cylinder
from marxs.optics import FlatDetector, CATGrating
from marxs.simulator import Propagator
from marxs.missions.mitsnl.catgrating import CATL1L2Stack, catsupportbars, l1_order_selector
from marxs.design import tolerancing as tol
from marxs.design.bendgratings import NumericalChirpFinder, chirp_gratings

from . import ralfgrating
from .mirror import MetaShell, MetaShellAperture



from marxs.base import TagVersion

tagversion = TagVersion(SATELLIT='LYNX')

filterdata = Table.read(get_pkg_data_filename('data/filtersqe.dat'),
                        format='ascii.ecsv')
filterqe = optics.GlobalEnergyFilter(filterfunc=interp1d(filterdata['energy'] / 1000,
                                                         filterdata['Total_throughput']))

conf = {'metashellgeometry': Table.read(get_pkg_data_filename('data/metashellgeom.dat'),
                                        format='ascii.ecsv'),
        'metashelleff': Table.read(get_pkg_data_filename('data/metashelleff.dat'),
                                    format='ascii.ecsv'),
        'inplanescatter': 2e-6 * u.rad,
        'perpplanescatter': 2e-6 * u.rad,
        'blazeang': np.deg2rad(1.6),
        'focallength': 10000.,
        'design_tilted_torus': 9.6e3, 
        'alphafac': 2.2,
        'betafac': 4.4,
        'grating_size': np.array([20., 50.]),
        'grating_frame': 2.,
        'det_kwargs': {'y_range': [400, 800],
                       'd_element': [16.884, 16.884],
                       'elem_class': optics.FlatDetector,
                       'elem_args': {'zoom': [1, 8.192, 8.192],
                                     'pixsize': 0.016}},
        'gas_kwargs': {'parallel_spec': np.array([0., 1., 0., 0.]),
                       'normal_spec': np.array([0, 0, 0, 1]),
                       'x_range': [7e3, 1e4],
                       'elem_class': CATL1L2Stack,
                       'elem_args': {'d': 2e-4,
                                     'order_selector': ralfgrating.order_selector_Si,
                                     'l1_dims': {'bardepth': 5.7 * u.micrometer,
                                                 'period': 0.005 * u.mm,
                                                 'barwidth': 0.0005 * u.mm},
                                     'l2_dims': {'bardepth': 0.5 * u.mm,
                                                 'period': 0.966 * u.mm,
                                                 'barwidth': 0.05 * u.mm},
                                     # Set sigma to 0. Effectively disables this
                                     # factor.
                                     'qualityfactor': {'d': 200. * u.um,
                                                       'sigma': 0 * u.um},
                                     'l1_order_selector': l1_order_selector,
                                     },
                       },
    }


def add_rowland_to_conf(conf):
    conf['alpha'] = conf['alphafac'] * conf['blazeang']
    conf['beta'] = conf['betafac'] * conf['blazeang']
    R, r, pos4d = design_tilted_torus(conf['design_tilted_torus'], conf['alpha'], conf['beta'])
    conf['rowland'] = RowlandTorus(R, r, pos4d=pos4d)
    conf['blazemat'] = transforms3d.axangles.axangle2mat(np.array([0, 0, 1]),
                                                         -conf['blazeang'])

add_rowland_to_conf(conf)

conf_5050 = copy.copy(conf)
conf_5050['grating_size'] = np.array([50., 50.])

conf_chirp = copy.copy(conf)
conf_chirp['grating_size'] = np.array([80., 160.])
conf_chirp['chirp_energy'] = 0.6
# Use of a non-integer order makes no sense physically
# but is just a numerical tool to optimize at the blaze peak
conf_chirp['chirp_order'] = -5.4


class LynxGAS(CircularMeshGrid):
    def __init__(self, conf):
        gg = conf['gas_kwargs']
        # Add in / update keywords that are calculated based on other conf
        # entries
        gg['rowland'] = conf['rowland']
        gg['d_element'] = conf['grating_size'] + 2 * conf['grating_frame']
        gg['radius'] = [np.min(conf['metashellgeometry']['r_inner']),
                        np.max(conf['metashellgeometry']['r_outer'])]
        gg['elem_args']['zoom'] = [1.,
                                   conf['grating_size'][0] / 2.,
                                   conf['grating_size'][1] / 2.]
        gg['elem_args']['orientation'] = conf['blazemat']

        super().__init__(**gg)

        # Color gratings according to the sector they are in
        for e in self.elements:
            e.display = copy.deepcopy(e.display)
            # Angle from baseline in range 0..pi
            ang = np.arctan2(e.pos4d[1, 3], e.pos4d[2, 3]) % (np.pi)
            # pick one colors
            e.display['color'] = 'rgb'[int(ang / np.pi * 3)]


class RowlandDetArray(design.rowland.RectangularGrid):
    def __init__(self, conf):
        super(RowlandDetArray, self).__init__(conf['rowland'], **conf['det_kwargs'],
                                              guess_distance=25.)


# Place an additional detector on the Rowland circle.
detcirc = optics.CircularDetector(geometry=Cylinder.from_rowland(conf['rowland'],
                                                                 width=50,
                                                                 rotation=np.pi,
                                                                 kwargs={'phi_lim':[-np.pi/2, np.pi/2]}))
detcirc.loc_coos_name = ['detcirc_phi', 'detcirc_y']
detcirc.detpix_name = ['detcircpix_x', 'detcircpix_y']

flatdet = FlatDetector(zoom=[1, 1e5, 1e5])
flatdet.loc_coos_name = ['detinf_x', 'detinf_y']
flatdet.detpix_name = ['detinfpix_x', 'detinfpix_y']
flatdet.display['shape'] = 'None'


class PerfectLynx(simulator.Sequence):
    '''Default Definition of Lynx without any misalignments'''
    def add_mirror(self, conf):
        return [MetaShellAperture(conf), MetaShell(conf)]

    def add_gas(self, conf):
        return [LynxGAS(conf),
                catsupportbars]

    def add_detectors(self, conf):
        '''Add detectors to the element list

        This is a separate function that is called from __init__ because all
        detectors need different parameters. Placing this specific code in its own
        function makes it easy to override for derived classes.
        '''
        microcal = FlatDetector(zoom=[1, 50, 50])
        microcal.loc_coos_name = ['microcal_x', 'microcal_y']
        microcal.detpix_name = ['microcalpix_x', 'microcalpix_y']
        proj2 = analysis.ProjectOntoPlane()
        proj2.loc_coos_name = ['projcirc_y', 'projcirc_z']
        return [RowlandDetArray(conf),
                analysis.ProjectOntoPlane(),
                Propagator(distance=-1000),
                detcirc,
                proj2,
                Propagator(distance=-1000),
                flatdet,
                Propagator(distance=-1000),
                microcal]

    def post_process(self):
        self.KeepPos = simulator.KeepCol('pos')
        return [self.KeepPos]

    def __init__(self, conf=conf, **kwargs):
        elem = self.add_mirror(conf)
        elem.extend(self.add_gas(conf))
        elem.append(filterqe)
        elem.extend(self.add_detectors(conf))
        elem.append(tagversion)

        super(PerfectLynx, self).__init__(elements=elem,
                                          postprocess_steps=self.post_process(),
                                          **kwargs)
        if ('chirp_energy' in conf) and ('chirp_order' in conf):
            gratings = self.elements_of_class(CATGrating, subclass_ok=False)
            opt = NumericalChirpFinder(detcirc,
                                       order=conf['chirp_order'],
                                       energy=conf['chirp_energy'],
                                       d=conf['gas_kwargs']['elem_args']['d'])
            chirp_gratings(gratings, opt, conf['gas_kwargs']['elem_args']['d'])


class LynxForPlot(PerfectLynx):
    def add_detectors(self, conf):
        microcal = FlatDetector(zoom=[1, 50, 50])
        microcal.loc_coos_name = ['microcal_x', 'microcal_y']
        microcal.detpix_name = ['microcalpix_x', 'microcalpix_y']
        return [RowlandDetArray(conf),
                microcal]


class Lynx(PerfectLynx):
    def __init__(self, conf=conf, **kwargs):
        super().__init__(conf=conf, **kwargs)
        for row in conf['alignmentbudget']:
            elem = self.elements_of_class(row[0])
            if row[1] == 'global':
                tol.moveglobal(elem, *row[3])
            elif row[1] == 'individual':
                tol.wiggle(elem, *row[3])
            else:
                raise NotImplementedError('Alignment error {} not implemented'.format(row[1]))


def reformat_errorbudget(budget, globalfac=0.8):
    '''Reformat the error budget

    Last, global misalignment (that's not random) must be
    scaled in some way. Here, I use 0.8 sigma, which is the
    mean absolute deviation for a Gaussian.

    Parameters
    ----------
    budget : list
        See reference implementation for list format
    globalfac : ``None`` or float
        Factor to apply for global tolerances. A "global" tolerance is drawn
        only once per simulation. In contrast, for "individual" tolerances
        many draws are done and thus the resulting layout actually
        represents a distribution. For a "global" tolerance, the result hinges
        essentially on a single random draw. If this is set to ``None``,
        misalignments are drawn statistically. Instead, the tolerances can be
        scaled determinisitically, e.g. by "0.8 sigma" (the mean absolute
        deviation for a Gaussian distribution).
    '''
    for row in budget:
        tol = np.array(row[2], dtype=float)
        if row[1] == 'global':
            if globalfac is not None:
                tol *= globalfac
            else:
                tol *= np.random.randn(len(tol))

        row[3] = tol


align_requirement_in = [
    [LynxGAS, 'global', [1, 0.5, .5,
                         np.deg2rad(0.01), np.deg2rad(0.01), np.deg2rad(0.01)],
     None, 'CAT structure to mirrors'],
    [LynxGAS, 'individual', [.1, .25, .25,
                             np.deg2rad(0.1), np.deg2rad(0.2), np.deg2rad(0.1)],
     None, 'individual CAT gratings structure'],
    [RowlandDetArray, 'global', [0.1, 1., 3.,
                                 np.deg2rad(0.5), np.deg2rad(1), np.deg2rad(.01)],
     None, 'Camera to front assembly']]

align_requirement = copy.deepcopy(align_requirement_in)
reformat_errorbudget(align_requirement)

conf['alignmentbudget'] = align_requirement



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import astropy.units as u
from .. import Arcus
from ..defaults import DefaultSource, DefaultPointing
from marxs.missions.arcus.utils import config as arcusconf

import pytest

@pytest.mark.skipif('data' not in arcusconf,
                    reason='Test requires Arcus CALDB')
def test_blaze():
    '''Check that photons have a reasonable blaze. We do not want to hardcode a
    particular value here but if they are all zero or they are all over the
    place then something is probably wrong.
    '''
    s = DefaultSource()
    point = DefaultPointing()
    p = s.generate_photons(10000 * u.s)
    p = point(p)
    arc = Arcus()
    p = arc(p)
    ind = np.isfinite(p['blaze']) & (p['probability'] > 0)

    assert np.all(np.rad2deg(p['blaze'][ind]) > 1.)
    assert np.all(np.rad2deg(p['blaze'][ind]) < 3.)
    assert np.all(np.std(np.rad2deg(p['blaze'][ind])) < .2)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from numpy.testing import assert_allclose
from scipy.stats import kstest
import transforms3d
import pytest
from astropy.coordinates import SkyCoord
import astropy.units as u
from astropy.utils.data import get_pkg_data_filename

from ..rowland import (RowlandTorus, GratingArrayStructure,
                       RectangularGrid,
                       find_radius_of_photon_shell, design_tilted_torus,
                       CircularMeshGrid,
                       double_rowland_from_channel_distance,
                       add_offset_double_rowland_channels,
                       )
from ...optics.base import FlatOpticalElement
from ...source import PointSource, FixedPointing
from ...optics import MarxMirror, OrderSelector, FlatGrating
from ...math.utils import h2e, xyz2zxy

class mock_facet(FlatOpticalElement):
    '''Lightweight class with no functionality for tests.'''
    pass

def test_radius_of_photon_shell():
    mysource = PointSource(coords=SkyCoord(30., 30., unit="deg"))
    photons = mysource.generate_photons(1. * u.ks)
    mypointing = FixedPointing(coords=SkyCoord(30, 30., unit='deg'))
    photons = mypointing.process_photons(photons)
    marxm = MarxMirror(get_pkg_data_filename('hrma.par', package='marxs.optics'), position=np.array([0., 0, 0]))
    photons = marxm(photons)
    r1, r2 = find_radius_of_photon_shell(photons, 1, 9e3)
    assert abs(r1 - 433.) < 1.
    assert abs(r2 - 442.) < 1.
    r1, r2 = find_radius_of_photon_shell(photons, 1, 9e3, percentile=[49, 49.001])
    assert (r2 - r1) < 0.01

def test_design_tilted_torus():
    '''Test the trivial case with analytic answers and consistency for other angles'''
    R, r, pos4d = design_tilted_torus(10, 0., 0.)
    assert R == 5
    assert r == 5
    assert np.all(pos4d == np.eye(4))
    # This is a solution that looks good by hand, but could still be slightly wrong...
    R, r, pos4d = design_tilted_torus(10, np.deg2rad(3), np.deg2rad(6))
    assert r == pytest.approx(5.0068617299896054)
    assert R == pytest.approx(4.9794336175561327)
    assert pos4d[0,3] == pytest.approx(0.027390523158633273)
    # The z axis should be unchanged
    assert np.allclose(pos4d[[0,1],2], 0)

def test_design_tilted_torus_negative_angles():
    '''Tilt the torus the other way. Should be same geometry in general.'''
    R, r, pos4d = design_tilted_torus(10, np.deg2rad(3), np.deg2rad(6))
    Rn, rn, pos4dn = design_tilted_torus(10, np.deg2rad(-3), np.deg2rad(-6))
    assert R == pytest.approx(Rn)
    assert r == pytest.approx(rn)
    assert pos4d[0,3] == pytest.approx(pos4dn[0, 3])
    assert pos4d[2, 3] == 0 # torus in xy plane
    assert pos4dn[2, 3] == 0
    assert pos4d[1, 3] == pytest.approx(-pos4dn[1, 3])

def test_torus():
    '''Compare parametric equation of torus with non-parametric.
    '''
    R = 2.
    r = 1.

    angle = np.arange(0, 2 * np.pi, 0.1)
    phi, theta = np.meshgrid(angle, angle)
    phi = phi.flatten()
    theta = theta.flatten()
    mytorus = RowlandTorus(R=R, r=r)

    xyzw = mytorus.parametric(theta, phi)
    xyz = h2e(xyzw)

    assert np.allclose(mytorus.quartic(xyz), 0.)
    # This torus has pos4d=eye(4), so the same results should come out with this shortcut
    assert np.allclose(mytorus.quartic(xyz, transform=False), 0.)

def test_torus_solve_quartic():
    '''solve_quartic helps to find points on the torus.

    This test could make use of the hypothesis testing package.
    '''
    mytorus = RowlandTorus(5., 4.)
    xyzw = mytorus.parametric([.34, 1.23], [.4, 4.5])

    for i in [0, 1]:
        # generate random direction in homogeneous coordinates
        direction = np.random.rand(4) - 0.5
        direction[3] = 0
        assert np.allclose(xyzw[i, :],
                            mytorus.solve_quartic(origin=xyzw[i, :] + 0.1 * direction,
                                                  v=direction))


def test_rotated_torus():
    '''Test the torus equation for a set of points.

    Importantly, we use a parametric description of the torus here, which is
    different from the actual implementation.
    '''
    def parametrictorus(R, r, theta, phi):
        '''Just another way to specify a torus with y-axis as symmetry'''
        x = (R + r * np.cos(theta)) * np.cos(phi)
        z = (R + r * np.cos(theta)) * np.sin(phi)
        y = r * np.sin(theta)
        return np.array([x, y, z]).T

    angle = np.arange(0, 2 * np.pi, 0.1)
    phi, theta = np.meshgrid(angle, angle)
    phi = phi.flatten()
    theta = theta.flatten()
    xyz = parametrictorus(2., 1., theta, phi)
    # rotate -90 deg around y axis:
    #   x -> z
    #   z -> -x
    rotatedxyz = np.zeros_like(xyz)
    rotatedxyz[:, 0] = -xyz[:, 2]
    rotatedxyz[:, 1] = xyz[:, 1]
    rotatedxyz[:, 2] = xyz[:, 0]
    mytorus = RowlandTorus(R=2., r=1., orientation=transforms3d.axangles.axangle2mat([0,1,0], -np.pi/2))

    assert np.allclose(mytorus.quartic(rotatedxyz), 0.)

    assert np.allclose(rotatedxyz, h2e(mytorus.parametric(theta, phi)))


def test_torus_xyzw2parametric():
    '''Theta, phi coordinates for torus should round-trip to xyz'''
    R = 2.
    r = 1.

    angle = np.arange(0, 2 * np.pi, 0.1)
    phi, theta = np.meshgrid(angle, angle)
    phi = phi.flatten()
    theta = theta.flatten()
    mytorus = RowlandTorus(R=R, r=r)

    xyzw = mytorus.parametric(theta, phi)
    t, p = mytorus.xyzw2parametric(xyzw)
    assert np.allclose(np.mod(theta, 2. * np.pi), np.mod(t, 2. * np.pi))
    assert np.allclose(np.mod(phi, 2. * np.pi), np.mod(p, 2. * np.pi))


def test_torus_normal():
    '''Compare the analytic normal with a numeric result.

    The test computes a vector along the torus for small d_theta
    and d_phi. Both of those should be perpendicular to the normal.
    '''
    R = 2.
    r = 1.

    angle = np.arange(0, 2 * np.pi, 0.1)
    phi, theta = np.meshgrid(angle, angle)
    phi = phi.flatten()
    theta = theta.flatten()
    mytorus = RowlandTorus(R=R, r=r)

    xyz = mytorus.parametric(theta, phi)
    t1 = mytorus.parametric(theta + 0.00001, phi)
    p1 = mytorus.parametric(theta, phi + 0.00001)
    t0 = mytorus.parametric(theta - 0.00001, phi)
    p0 = mytorus.parametric(theta, phi - 0.00001)

    vec_normal = h2e(mytorus.normal(xyz))

    vec_delta_theta = h2e(t1) - h2e(t0)
    vec_delta_theta = vec_delta_theta  / np.linalg.norm(vec_delta_theta, axis=1)[:, None]
    vec_delta_phi = h2e(p1) - h2e(p0)
    vec_delta_phi = vec_delta_phi  / np.linalg.norm(vec_delta_phi, axis=1)[:, None]

    assert np.allclose(np.sqrt(np.sum(vec_normal * vec_normal, axis=1)), 1.)
    assert np.allclose(np.einsum('ij,ij->i', vec_normal, vec_delta_theta), 0.)
    assert np.allclose(np.einsum('ij,ij->i', vec_normal, vec_delta_phi), 0.)

def torus_cut_is_circle():
    '''Cutting the torus in a plane that includes the symmetry axis should give a cirlce.

    Check that that's true even for a tilted torus.
    '''
    R, r, pos4d = design_tilted_torus(10, np.deg2rad(-3), np.deg2rad(-6))
    torus = RowlandTorus(R, r, pos4d=pos4d)

    z = np.arange(0., torus.r *0.8, 5)
    x = np.zeros_like(z)
    xguess = torus.r + torus.R
    for i, iz in enumerate(z):
        x[i] = torus.solve_quartic(None, 0, iz, [0.99 * xguess, 1.1 * xguess])
        xguess = x[i]

    # This should be a circle of radius r around a center.
    # Where exactly is the center?
    # Assuming no y rotation:
    circle_center = np.dot(torus.pos4d[:3,:3], torus.R * np.array([1, 0,0])) + torus.pos4d[:3,3]
    circle_center_xz = circle_center[[0,2]]
    p = np.vstack([x,z])
    d = np.linalg.norm(p - circle_center_xz[:, np.newaxis], axis=0)
    assert np.allclose(d, torus.r)

def test_GratingArrayStructure():
    '''Check a GAS for consistency

    The expected numbers for this one are hand-checked and plotted and thus
    can be taken as known-good to prevent regressions later.
    '''
    myrowland = RowlandTorus(10000., 10000.)
    gas = GratingArrayStructure(rowland=myrowland,
                                d_element=[30., 30.],
                                radius=[300., 600.],
                                phi=[-0.2*np.pi, 0.2*np.pi], elem_class=mock_facet)
    assert gas.max_elements_on_arc(300.) == 12
    angles = gas.distribute_elements_on_arc(315.) % (2. * np.pi)
    # This is a wrap-around case. Hard to test in general, but here I know the numbers
    assert np.alltrue((angles < 0.2 * np.pi) | (angles > 1.8 * np.pi))
    assert gas.max_elements_on_radius(gas.radius) == 10
    assert np.all(gas.distribute_elements_on_radius() == np.arange(315., 600., 30.))
    assert len(gas.elem_pos) == 177
    # fig, axes = plt.subplots(2,2)
    # for elem in [[axes[0, 0], 0, 1], [axes[0, 1], 0, 2], [axes[1, 0], 1, 2]]:
    #     for f in gas.elem_pos:
    #         elem[0].plot(f[elem[1],3], f[elem[2],3], 's')
    # plt.show()  # or move mouse in window or something.

    # There is a test missing here that the rotational part works correctly.
    # I just cannot think of a good way to check that right now.

    # Check that initially all uncertainties are 0
    for i in range(len(gas.elem_pos)):
        assert np.allclose(gas.elem_pos[i], gas.elements[i].pos4d)

def test_GAS_multipleradii():
    '''Radius can be a list of many pairs.'''
    myrowland = RowlandTorus(1000., 1000.)
    gas1 = GratingArrayStructure(rowland=myrowland, d_element=[60., 60.],
                                 radius=[300., 400.], elem_class=mock_facet)
    gas2 = GratingArrayStructure(rowland=myrowland, d_element=[60., 60.],
                                 radius=[500., 540.], elem_class=mock_facet)
    gas = GratingArrayStructure(rowland=myrowland, d_element=[60., 60.],
                                radius=[300., 400., 500., 540.],
                                elem_class=mock_facet)

    r1 = gas1.distribute_elements_on_radius()
    r2 = gas2.distribute_elements_on_radius()
    r = gas.distribute_elements_on_radius()
    assert np.all(np.in1d(r, np.union1d(r1, r2)))


def test_GratingArrayStructure_2pi():
    '''test that delta_phi = 2 pi means "full circle" and not 0
    '''
    myrowland = RowlandTorus(10000., 10000.)
    gas = GratingArrayStructure(rowland=myrowland, d_element=[30., 30.],
                                radius=[300., 600.],
                                phi=[0, 2*np.pi], elem_class=mock_facet)
    assert gas.max_elements_on_arc(300.) > 10
    n = len(gas.elem_pos)
    yz = np.empty((n, 2))
    for i, p in enumerate(gas.elem_pos):
        yz[i, :] = p[1:3, 3]
    phi = np.arctan2(yz[:, 1], yz[:, 0])
    ks, pvalue = kstest((phi + np.pi) / (2 * np.pi), 'uniform')
    assert pvalue > 0.3  # It's not exactly uniform because of finite size of elements.

def test_GAS_facets_on_radius():
    '''test distribution of elements on radius for d_r is non-integer multiple of d_element.'''
    myrowland = RowlandTorus(1000., 1000.)
    gas = GratingArrayStructure(rowland=myrowland, d_element=[60., 60.],
                                radius=[300., 400.], elem_class=mock_facet)
    assert np.all(gas.distribute_elements_on_radius() == [320., 380.])
    gas.radius = [300., 340.]
    assert gas.distribute_elements_on_radius() == [320.]

def test_facet_rotation_via_facetargs():
    '''The numbers for the blaze are not realistic.'''
    gratingeff = OrderSelector(np.arange(-3, 4))
    mytorus = RowlandTorus(9e3/2, 9e3/2)
    mygas = GratingArrayStructure(rowland=mytorus, d_element=[60., 60.],
                                  radius=[538., 550.], elem_class=FlatGrating,
                                  elem_args={'zoom': 30, 'd':0.0002, 'order_selector': gratingeff})
    blaze = transforms3d.axangles.axangle2mat(np.array([0,1,0]), np.deg2rad(15.))
    mygascat = GratingArrayStructure(rowland=mytorus, d_element=[60., 60.],
                                     radius=[538., 550.], elem_class=FlatGrating,
                                     elem_args={'zoom': 30, 'orientation': blaze, 'd':0.0002, 'order_selector': gratingeff})
    assert np.allclose(np.rad2deg(np.arccos(np.dot(mygas.elements[0].geometry['e_x'][:3], mygascat.elements[0].geometry['e_x'][:3]))), 15.)

def test_persistent_facetargs():
    '''Make sure that facet_args is still intact after generating facets.

    This is important to allow tweaks of a single parameter and then to regenerate the facets.
    '''
    gratingeff = OrderSelector(np.arange(-3, 4))
    mytorus = RowlandTorus(9e4/2, 9e4/2)
    # id_col is automatically added in GAS is not present here.
    # So, pass in an id_col to make sure the comparison below will still work.
    facet_args = {'zoom': 30, 'd':0.0002, 'order_selector': gratingeff, 'id_col': 'facet'}
    mygas = GratingArrayStructure(rowland=mytorus, d_element=[60., 60.],
                                  radius=[5380., 5500.], elem_class=FlatGrating,
                                  elem_args=facet_args)
    assert mygas.elem_args == facet_args

def test_run_photons_through_gas():
    '''And check that they have the expected labels.

    No need to check here that the grating equation works - that's part of the grating tests.
    '''
    # Setup only.
    mysource = PointSource(coords=SkyCoord(30., 30., unit="deg"))
    photons = mysource.generate_photons(1. * u.ks)
    mypointing = FixedPointing(coords=SkyCoord(30, 30., unit='deg'))
    photons = mypointing.process_photons(photons)
    marxm = MarxMirror(get_pkg_data_filename('hrma.par', package='marxs.optics'), position=np.array([0., 0, 0]))
    photons = marxm(photons)
    gratingeff = OrderSelector([-1, 0, 1])
    facet_args = {'zoom': 30, 'd':0.0002, 'order_selector': gratingeff}
    mytorus = RowlandTorus(9e3/2, 9e3/2)

    # Now the real test with different input for facet_args

    for i in [1, 2, 3, 4]:
        if i == 1:
            kwargs = {}
            f = 'facet'
        elif i == 2:
            kwargs = {'id_col': 'qewr'}
            f = 'qewr'
        elif i == 3:
            facet_args['id_col'] = 'ttt'
            f = 'ttt'
            kwargs = {}
        elif i == 4:
            facet_args['id_col'] = 'uuu'
            f = 'uuu'
            kwargs = {'id_col': 'yyy'}

        mygas = GratingArrayStructure(rowland=mytorus, d_element=[60., 60.],
                                      radius=[538., 550.], elem_class=FlatGrating,
                                      elem_args=facet_args, **kwargs)

        p = mygas(photons.copy())
        indorder = np.isfinite(p['order'])
        indfacet = p[f] >=0
        assert np.all(indorder == indfacet)

        assert set(p['order'][indorder]) == set([-1, 0, 1])
        # -1 means no hit - passing between facets
        allfacets = set([-1]).union(set(np.arange(len(mygas.elements))))
        assert set(p[f]).issubset(allfacets)


def test_RectangularGrid_rotated():
    '''Test an array with different rotations.

    In this case, we rotate the Rowland torus by -30 deg and then
    check that the e_z vector is rotated 30 deg with respect to the
    coordinate system.
    '''
    pos4d = transforms3d.axangles.axangle2aff([1, 0, 0], np.deg2rad(-30))
    myrowland = RowlandTorus(10000., 10000., pos4d=pos4d)
    # Along the way we normally would orient the detector.
    ccds = RectangularGrid(rowland=myrowland, d_element=[30., 30.],
                           y_range=[-100, 100],
                           elem_class=mock_facet,
                           guess_distance=100)
    assert len(ccds.elements) == 7
    for e in ccds.elements:
        assert np.isclose(np.dot([0, -0.8660254, 0.5], e.geometry['e_z'][:3]), 0, atol=1e-4)


def test_RectangularGridone():
    '''Test an array of one element in default position'''
    myrowland = RowlandTorus(10000., 10000.)
    # Along the way we normally would orient the detector.
    ccds = RectangularGrid(rowland=myrowland, d_element=[300., 300.],
                           y_range=[-1, 1],
                           elem_class=mock_facet,
                           guess_distance=25)
    assert len(ccds.elements) == 1
    # multiply because of pointing inward vs. outward
    assert np.allclose(ccds.elements[0].pos4d, np.eye(4))


def test_RegtangularGrid_offcenter_unaffected():
    '''Make sure the coordinate system of the xyz_range is in global coordinates.
    In this test, we check the z position, which in unchanged because the torus
    center is moved only in the y direction.'''
    R, r, pos4d = design_tilted_torus(5000, 0.07, 0.15)
    rowland = RowlandTorus(R, r, pos4d=pos4d)
    ccds = RectangularGrid(rowland=rowland, d_element=[5., 5.],
                           z_range=[10, 20],
                           y_range=[600, 615],
                           elem_class=mock_facet,
                           guess_distance=25)

    for e in ccds.elements:
        assert (e.geometry['center'][2] > 10) and (e.geometry['center'][2] < 20)


def test_RegtangularGrid_offcenter_affected():
    '''Make sure the coordinate system of the xyz_range is in global coordinates.
    In this test, we check the y position.'''
    R, r, pos4d = design_tilted_torus(50000, 0.007, 0.015)
    rowland = RowlandTorus(R, r, pos4d=pos4d)
    ccds = RectangularGrid(rowland=rowland, d_element=[50., 50.],
                           y_range=[600, 800],
                           elem_class=mock_facet,
                           guess_distance=25)
    y_offset = rowland.geometry['center'][1]
    for e in ccds.elements:
        assert (e.geometry['center'][1] > 600 + y_offset) and \
               (e.geometry['center'][1] < 800 + y_offset)



def test_nojumpsinCCDorientation():
    '''Regression test: CCD orientation should change smoothly along the circle.'''
    R, r, pos4d = design_tilted_torus(12e3, 0.07, 0.15)
    rowland = RowlandTorus(R, r, pos4d=pos4d)
    det = RectangularGrid(rowland=rowland,
                             elem_class=mock_facet,
                             d_element=[49.652, 49.652],
                             y_range=[-200, +200],
                             guess_distance=25.)

    # If all is right, this will vary very smoothly along the circle.
    xcomponent = np.array([e.geometry['e_y'][0] for e in det.elements])
    xcompdiff = np.diff(xcomponent)
    assert (np.max(np.abs(xcompdiff)) / np.median(np.abs(xcompdiff))) < 1.1


def test_circularMeshGrid():
    '''Check MeshGrid against a hand-checked and thus known-good solution'''
    myrowland = RowlandTorus(10000., 10000.)
    gas = CircularMeshGrid(rowland=myrowland,
                           radius=(100., 300.),
                           d_element=(30., 50),
                           elem_class=mock_facet,
                           optimize_axis=np.array([1, 0, 0, 0]),
                           parallel_spec=np.array([0., 1., 0., 0.]),
                           normal_spec=np.array([1, 0, 0, 1]))
    assert len(gas.elements) == 178
    poslist = np.stack(gas.elem_pos)
    assert np.all(poslist[:, 0, 3] > 19000)
    assert np.all(poslist[:, 0, 3] < 20000)
    assert np.all(np.abs(poslist[:, 1, 3] < 300))
    assert np.all(np.abs(poslist[:, 2, 3] < 300))

def test_circularMeshGrid_rotated():
    '''Check MeshGrid against a hand-checked and thus known-good solution.
    Repeats previous test but for a different axes'''
    myrowland = RowlandTorus(10000., 10000.)
    myrowland.pos4d = xyz2zxy @ myrowland.pos4d
    gas = CircularMeshGrid(rowland=myrowland,
                           radius=(100., 300.),
                           d_element=(30., 50),
                           elem_class=mock_facet,
                           optimize_axis=np.array([0, 0, 1, 0]),
                           parallel_spec=np.array([0., 1., 0., 0.]),
                           normal_spec=np.array([0, 1, 0, 1]))
    assert len(gas.elements) == 178
    poslist = np.stack(gas.elem_pos)
    assert np.all(poslist[:, 2, 3] > 19000)
    assert np.all(poslist[:, 2, 3] < 20000)
    assert np.all(np.abs(poslist[:, 0, 3] < 300))
    assert np.all(np.abs(poslist[:, 1, 3] < 300))


def test_double_rowland_from_channel_distance():
    '''Check that the double torus is in the right position.
    The numbers are hand-checked and thus known-good.
    '''
    geom = double_rowland_from_channel_distance(600., 6000, 12000)
    assert geom['pos_opt_ax']['1'] == pytest.approx([0, -300, 0, 1])
    assert geom['pos_opt_ax']['1m'] == pytest.approx([0, 300, 0, 1])

    assert geom['rowland_detector'].R == pytest.approx(6000)
    assert geom['rowland_detector'].r == pytest.approx(6007.495318350236)
    assert geom['rowland_central'].geometry['center'] == pytest.approx([7.48596673, 599.62570166, 0.,  1.])


def test_offset_double_rowland_channels():
    '''Split the double torus into two and offset all four channels.

    Checked against the version of the code that's been in use for years
    for Arcus.
    '''
    geom = double_rowland_from_channel_distance(600., 6000, 12000)
    add_offset_double_rowland_channels(geom,
                                   offsets={'1': [0, -2.5, -7.5],
                                            '1m': [0, -2.5, -2.5],
                                            '2': [0, 2.5, 2.5],
                                            '2m': [0, 2.5, 12.5],
                                            })
    assert geom['pos_opt_ax']['2m'] == pytest.approx([0, 302.5, 12.5, 1. ])
    assert geom['rowland_1'].geometry['center'] == pytest.approx([7.48596673, 297.12570166, -7.5, 1.])


def test_rowlands_are_oriented_the_same_way():
    """When there are two rowland tori, they should be oriented the same way."""
    double = double_rowland_from_channel_distance(100, 500, 1000.0)
    add_offset_double_rowland_channels(
        double, offsets={"1": [0, -5, 0], "1m": [0, +5, 0]}
    )
    assert_allclose(
        double["rowland_central"].parametric(0, 0),
        double["rowland_central_m"].parametric(0, 0),
    )
    # The "top" points in the same direction. Of course, they are offset from each other
    # so the numbers are not exactly the same.
    assert_allclose(double["rowland_1"].parametric(0, 0), [1000.0, -55.0, 0.0, 1.0])
    assert_allclose(double["rowland_1m"].parametric(0, 0), [1000.0, 55.0, 0.0, 1.0])


import numpy as np
from astropy.table import Table
import astropy.units as u
import warnings

__all__ = ['SimulationSetupWarning', 'generate_test_photons',
           'DataFileFormatException', 'tablerows_to_2d']


class SimulationSetupWarning(Warning):
    pass

warnings.filterwarnings("always", ".*", SimulationSetupWarning)


def generate_test_photons(n=1):
    '''Generate a photon structure for testing.

    This function returns a table of ``n`` identical photons with
    standard properties:

    - position: ``x=1; y=0; z=0``
    - direction: negative x-axis
    - energy: 1 keV
    - polarization: y-direction

    This is useful for testing purposes.

    Parameters
    ----------
    n : int
        number of photons to generate

    Returns
    -------
    photons : `astropy.table.Table`
        Table of ``n`` identical photons.
    '''
    dir = np.tile(np.array([-1., 0., 0., 0.]), (n, 1))
    pos = np.tile(np.array([1., 0., 0., 1.]), (n, 1))
    pol = np.tile(np.array([0., 1., 0., 0.]), (n, 1))
    photons = Table(
        {
            "pos": pos,
            "dir": dir,
            "energy": np.ones(n) * u.keV,
            "polarization": pol,
            "probability": np.ones(n),
        }
    )
    return photons


class DataFileFormatException(Exception):
    '''Exception for grating efficiency files not matching expected format.'''
    pass


def tablerows_to_2d(tab):
    '''Get a 2d array from an input table.

    In the table, the data is flattened to a 1d form.
    The first two columns are x and y, like this:
    The first column looks like this with many duplicates:
    [1,1,1,1,1,1,2,2,2,2,2,2,3,3,3, ...].
    Column B repeats like this: [1,2,3,4,5,6,1,2,3,4,5,6,1,2,3, ...].

    All remaining columns are data on the same x-y grid, and the grid
    has to be regular.

    Parameters
    ----------
    tab : `astropy.table.Table`
        Table as read in. Useful to access units or other meta data.

    Returns
    -------
    x, y : `astropy.table.Column`
        Unique entries in first and second column
    dat : np.array
        The remaining outputs are np.arrays of shape (len(x), len(y))
    '''
    x = tab.columns[0]
    y = tab.columns[1]
    n_x = len(set(x))
    n_y = len(set(y))
    if len(x) != (n_x * n_y):
        raise DataFileFormatException('Data is not on regular grid.')

    x_arr = tab.columns[0].data.reshape(n_x, n_y)
    y_arr = tab.columns[1].data.reshape(n_x, n_y)
    if not (np.allclose(x_arr, x_arr[:, 0][:, None]) and
            np.allclose(y_arr, y_arr[0, :][None, :])):
        raise DataFileFormatException('Input table x, y not sorted as expected.')

    x = x[::n_y]
    y = y[:n_y]
    coldat = [tab[d].data.reshape(n_x, n_y) for d in tab.columns[2:]]

    return x, y, coldat



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from transforms3d.euler import euler2mat
from transforms3d.affines import compose
import astropy.units as u
from astropy.utils.data import get_pkg_data_path
from astropy.table import Table
from scipy.interpolate import interp1d

from marxs.optics import PerfectLens, GlobalEnergyFilter, RadialMirrorScatter
from marxs.optics.aperture import CircleAperture
from marxs.simulator import Parallel, Sequence
from marxs.math.utils import xyz2zxy

__all__ = [
    "aperture",
    "spogeom",
    "spo_pos4d",
    "spogeom2pos4d",
    "SPOChannelMirror",
    "ScatterPerChannel",
    "geometricthroughput",
    "willingaleloss",
    "spomounting",
]

spogeom = Table.read(get_pkg_data_path('data/xous.csv'), format='ascii.ecsv')
spogeom['r_mid'] = (spogeom['outer_radius'] + spogeom['inner_radius']) / 2

def spogeom2pos4d(spogeom):
    spo_pos4d = []
    # Convert angle to quantity here to make sure that unit is taken into account
    for row, ang in zip(spogeom,
                        u.Quantity(spogeom['clocking_angle']).to(u.rad).value):
        spo_pos4d.append(compose([0,  # focallength,  # - spogeom[i]['d_from_12m']
                                  row['r_mid'] * np.sin(ang),
                                  row['r_mid'] * np.cos(ang)],
                                  euler2mat(-ang, 0., 0.),
                                  # In detail this should be (primary_length + gap + secondary_length) / 2
                                  # but the gap is somewhat complicated and this is only used
                                  # for display, we'll ignore that for now.
                                  [row['primary_length'],
                                   row['azwidth'] / 2.,
                                   # 36 reflective plates in each stack
                                   36 * row['plate_height'] / 2.
                                   ]
                                  )
                         )
    return spo_pos4d

spo_pos4d = [np.dot(xyz2zxy, s) for s in spogeom2pos4d(spogeom)]


aperture = CircleAperture(position=[0, 0, 12200],
                          zoom=[1,
                                np.max(spogeom['outer_radius']) * 1.05,
                                np.max(spogeom['outer_radius']) * 1.05],
                          orientation=xyz2zxy[:3, :3])


class SPOChannelMirror(Parallel):
    '''A collection of PerfectLensSegments into plates

    A number of SPOs are mounted together on a single plate and they are
    grouped into a Parallel structure here.

    Parameters
    ----------
    spo_pos4d : list of (4, 4) arrays
        pos4d arrays for each SPO or XOU pair
    spogeom : `astropy.table.Table`
        Table with columns "rmid" (radius from optical axis for each SPO / XOU pair) and
        "focal_length".
    reflectivity_interpolator : callable
        See `PerfectLensSegment`
    '''
    def __init__(self, **kwargs):
        kwargs['elem_pos'] = kwargs.pop('spo_pos4d')
        spogeom = kwargs.pop('spo_geom')
        kwargs["elem_class"] = PerfectLens
        kwargs['elem_args'] = {'d_center_optical_axis': list(spogeom['r_mid']),
                               'focallength': list(spogeom['focal_length']),
                               'reflectivity_interpolator': kwargs.pop('reflectivity_interpolator')}
        kwargs['id_col'] = 'xou'
        super().__init__(**kwargs)


geometricthroughput = GlobalEnergyFilter(filterfunc=lambda e: 0.92609,
                                         name='SPOgeometricthrougput')
'''Open area of each plate (i.e. pore width/pore pitch).

This is set by having a 0.17 mm wide rib and a 2.3 mm rib pitch.
'''

tab = Table.read(get_pkg_data_path('data/lossterm.csv'), format='ascii.ecsv')
willingaleloss = GlobalEnergyFilter(
    filterfunc=interp1d(tab['energy'].to(u.keV, equivalencies=u.spectral()),
                        tab[tab.colnames[1]]),
    name='Loss_from_nonideal_spos')
'''Additional loss terms from pore-level ray-tracing (tabluted).

Includes, e.g. the fact the ribs are parallel, not radial etc.
'''


class ScatterPerChannel(RadialMirrorScatter):
    '''A scatter of infinite size that identifies photons by spo id

    This bypasses the intersection calculation and instead
    just selects photons for this scatter by spo id.

    Parameters
    ----------
    min_id : integer
    max_id : integer
        Photons with xou id between ``min_id`` and ``max_id`` are
        scattered.
    '''
    display = {'shape': 'None'}
    loc_coos_name = ['scat_y', 'scat_z']

    def __init__(self, **kwargs):
        self.min_id = kwargs.pop('min_id')
        self.max_id = kwargs.pop('max_id')
        super().__init__(**kwargs)

    def __call__(self, photons):
        # interpos is used to automatically set new position
        # (which we want unaltered, thus we pass pos) and local coords
        # but intercoos we want to set to a useful number for analysis later
        intersect, interpos, intercoos = self.geometry.intersect(photons['dir'].data,
                                                                 photons['pos'].data)
        # intersect is done based on xou to avoid problem of overlap
        # between channels
        intersect = ((photons['xou'] >= self.min_id) &
                     (photons['xou'] < self.max_id))
        return self.process_photons(photons, intersect, photons['pos'].data,
                                    intercoos)


def spomounting(photons):
    '''Remove photons that do not go through an SPO but hit the
    frame part of the petal.'''
    photons['probability'][photons['xou'] < 0] = 0.
    return photons


class SimpleSPOs(Sequence):

    def __init__(self, conf, **kwargs):
        mirror = []
        entrancepos = [0, 0, 12000]
        mirror.append(SPOChannelMirror(position=entrancepos,
                                       reflectivity_interpolator=conf['reflectivity_interpolator'],
                                       spo_pos4d=conf['spo_pos4d'],
                                       spo_geom=conf['spo_geom']))
        mirror.append(ScatterPerChannel(position=entrancepos,
                                        min_id=0,
                                        max_id=10000,
                                        inplanescatter=conf['inplanescatter'],
                                        perpplanescatter=conf['perpplanescatter'],
                                        orientation=xyz2zxy[:3, :3]))
        mirror.append(spomounting)
        mirror.append(geometricthroughput)
        mirror.append(willingaleloss)
        super().__init__(elements=mirror, **kwargs)



'''MARXS supports several backends for 3D plotting; each of which is implemented as a submodule.

To plot using a specific backend, import the relevant submodule.
'''
from astropy import config as _config

class Conf(_config.ConfigNamespace):
    """
    Configuration parameters for my subpackage.
    """
    xyz_precision = _config.ConfigItem(
        1, 'Precision of coordinate output when floating point numbers are written to text, e.g. json. This is used as `np.round(xyz, p)` to reduce the size of text files.')
    color_precision = _config.ConfigItem(
        2, 'Precision for writing colors as tuples to text files, e.g. json. This is used as `np.round(color, p)` to reduce the size of text files.')

# Create an instance for the user
conf = Conf()


# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.stats import sigma_clipped_stats
import astropy.units as u

from ..optics import FlatDetector, CircularDetector, OrderSelector
from ..design import RowlandTorus
from . import (find_best_detector_position)
from .analysis import sigma_clipped_std
from ..math.geometry import Cylinder

__all__ = ['AnalysisError',
           'resolvingpower_per_order',
           'resolvingpower_from_photonlist',
           'resolvingpower_from_photonlist_robust',
           'weighted_per_order',
           'average_R_Aeff',
           'CaptureResAeff',
           'CaptureResAeff_CCDgaps',
           ]

class AnalysisError(Exception):
    pass


def resolvingpower_per_order(gratings, photons, orders, detector=None,
                             colname='det_x'):
    '''Calculate the resolving power in every grating order.

    As input this function takes a Grating Array Structure (``gratings``) and
    a list of photons ready to hit the grating, i.e. the photons have already
    passed through aperture and mirror in e.g. a Chandra-like design. The
    function will take the same input photons and send them through the gas
    looping over the grating orders. In each step of the loop all photons are
    send to the same order, thus the statistical uncertainty on the measured
    spectral resolving power (calculated as ``pos_x / FWHM_x``) is the same
    for every order and is set by the number of photons in the input list.

    As a side effect, the function that selects the grating orders for
    diffraction in ``gas`` will be changed. Pass a deep copy of the GAS if
    this could affect consecutive computations.

    Unlike `~marxs.analysis.gratings.resolvingpower_from_photonlist` this
    function ray-traces the input photon list from the gratings to the
    detectors.

    Parameters
    ----------
    gratings : `marxs.simulator.Parallel`
        Structure that holds individual grating elements,
        e.g. `marxs.design.GratingArrayStructure`
    photons : `astropy.table.Table`
        Photon list to be processed for every order
    orders : np.array of type int
        Order numbers
    detector : marxs optical element or `marxs.design.RowlandTorus` or
        ``None``. Photons are projected onto a detector. There are three ways
        to define this detector:

        - Pass in an instance of an optical element (e.g. a
          `marxs.optics.FlatDetector`).
        - Pass in a `marxs.design.RowlandTorus`. This function will generate
          a detector that follows the Rowland circle.
        - ``None``. A flat detector in the yz plane is used, but the x position
          for this detector is numerically optimized in each step.

    colname : string
        Name of the column that labels the dispersion direction. This is only
        used if a detector instance is passed in explicitly.

    Returns
    -------
    res : np.array
        Resolution defined as FWHM / position for each order.
    fwhm : np.array
        FWHM for each order
    info : dict
        Dictionary with more information, e.g. fit results
    '''
    res = np.zeros_like(orders, dtype=float)
    fwhm = np.zeros_like(orders, dtype=float)
    info = {}

    if detector is None:
        info['method'] = 'Detector position numerically optimized'
        info['fit_results'] = []
        col = 'det_x'  # 0 at center, detpix_x is 0 in corner.
        zeropos = 0.
    elif isinstance(detector, RowlandTorus):
        det = CircularDetector(geometry=Cylinder.from_rowland(detector,
                                                              width=1e5,
                                                              rotation=np.pi,
                                                              kwargs={'phi_lim':[-np.pi/2, np.pi/2]}))
        info['method'] = 'Circular detector on Rowland circle'
        col = 'detpix_x'
        # Find position of order 0.
        # if Rowland torus is tilted, might not be at phi=0.
        pg = photons.copy()
        pg = det(pg)
        # Remove photons that miss a mirror etc.
        pg = pg[pg['probability'] > 0.]
        zeropos, temp1, temp2 = sigma_clipped_stats(pg[col])
    else:
        det = detector
        info['method'] = 'User defined detector'
        col = colname
        # Find position of order 0.
        pg = photons.copy()
        pg = det(pg)
        zeropos, temp1, temp2 = sigma_clipped_stats(pg[col])

    for i, order in enumerate(orders):
        gratingeff = OrderSelector([order])
        gratings.elem_args['order_selector'] = gratingeff
        for elem in gratings.elements:
            elem.order_selector = gratingeff

        pg = photons.copy()
        pg = gratings(pg)
        # Remove photons that slip between the gratings
        pg = pg[(pg['order'] == order) & (pg['probability'] > 0.)]
        if detector is None:
            xbest = find_best_detector_position(pg, objective_func=sigma_clipped_std)
            info['fit_results'].append(xbest)
            det = FlatDetector(position=np.array([xbest.x, 0, 0]), zoom=1e5)
        pg = det(pg)
        meanpos, medianpos, stdpos = sigma_clipped_stats(pg[col])
        fwhm[i] = 2.3548 * stdpos
        res[i] = np.abs((meanpos - zeropos) / fwhm[i])
    return res, fwhm, info


def weighted_per_order(data, orders, energy, gratingeff):
    '''Summarize a per-order table of a quantity such as spectral resolution.

    `marxs.analysis.fwhm_per_order` produces a set of data for each grating
    order, most notably the spectral resolution achieved in every spectral
    order for every energy. In practice, however, most orders see only a very
    small number of photons and will not contribute significantly to the
    observed signal.

    This method provides one way to summarize the data by calculating the
    weighted mean of the resolution for each energy, weighted by the
    probability of photons to be diffracted into that order ( = the expected
    fraction).

    Parameters
    ----------
    data : (N, M) np.array
        Array with data for N orders and M energies, e.g. spectral resolution
    orders : np.array of length N of type int
        Order numbers
    energy : np.array for length M of type float
        Energies for each entry in resolution
    gratingeff : `marxs.optics.gratings.EfficiencyFile`
        Object that holds the probability that a photon on a certain energy is
        diffracted to a specific order.

    Returns
    -------
    weighted_res : np.array of length M
        Weighted resolution for each energy
    '''
    if len(orders) != data.shape[0]:
        raise ValueError('First dimension of "data" must match length of "orders".')
    if len(energy) != data.shape[1]:
        raise ValueError('Second dimension of "data" must match length of "energy".')

    weights = np.zeros_like(data)
    for i, o in enumerate(orders):
        ind_o = (gratingeff.orders == o).nonzero()[0]
        if len(ind_o) != 1:
            raise KeyError('No data for order {0} in gratingeff'.format(o))
        en_sort = np.argsort(gratingeff.energy)
        weights[i, :] = np.interp(energy, gratingeff.energy[en_sort],
                                  gratingeff.prob[:, ind_o[0]][en_sort])

    return np.ma.average(data, axis=0, weights=weights)


def average_R_Aeff(r, aeff, axis=None):
    '''Average R and Aeff over one dimension e.g. channel or orders

    Aeff is summed over the input, while R is calculated as the
    weighted average, using Aeff as the weight for each order.
    This function works for masked arrays and those with nan's.

    Unlike `weighted_per_order` this function summarizes measured
    quantities, so instead of an expected fraction that will be interpolated
    from some table, this uses the observed effective area as weight.

    Parameters
    ----------
    r : array
        Resolving power
    aeff : array
        Effective area
    axis : `None` or int
        Passed to `np.ma.average`. This can be used if the input arrays
        are multi-dimensional.

    Returns
    -------
    aeff_avg : array or float
        Summed effective area
    res_avg : array or float
        Average resolving power
    '''
    aeff_avg = aeff.sum(axis=axis)
    res_avg = np.ma.average(np.ma.masked_invalid(r),
                            weights=aeff, axis=axis)
    return res_avg, aeff_avg


def resolvingpower_from_photonlist(photons, orders,
                                   col='proj_x', zeropos=None,
                                   ordercol='order'):
    '''Calculate the resolving power for several grating orders

    If fewer than 20 photons are detected in a single order, this function
    returns nan values.

    Unlike `~marxs.analysis.gratings.resolvingpower_per_order` this method does
    not run any ray-trace simulations. It just extracts information from a
    photon list that is passed in.

    Parameters
    ----------
    photons : `astropy.table.Table`
        Photon event list
    orders : np.array
        Orders for which the resolving power will be calculated
    col : string
        Column name for the column holding the dispersion coordinate.
    zeropos : float or ``None``
        Value of column `col` where the zeroth order is found. If not given,
        this is calculated (assuming the zeroth order photons are part of the
        event list).
    ordercol : string
        Name of column that lists grating order for each photon

    Returns
    -------
    res : np.array
        resolving power for each order
    pos : np.array
        mean value of ``col`` for each order
    std : np.array
        standard deviation of the distribution of ``col`` for each order
    '''
    if zeropos is None:
        ind = (photons[ordercol] == 0)
        if ind.sum() < 20:
            raise AnalysisError('Too few photons in list to determine position of order 0 automatically.')
        zeropos, medzeropos, stdzero = sigma_clipped_stats(photons[col][ind])

    pos = np.zeros_like(orders, dtype=float)
    std = np.zeros_like(orders, dtype=float)

    for i, o in enumerate(orders):
        ind = (photons[ordercol] == o)
        if ind.sum() > 20:
            # There is a .value here to work around https://github.com/astropy/astropy/issues/13281
            meanpos, medianpos, stdpos = sigma_clipped_stats(photons[col][ind].value)
        else:
            meanpos, stdpos = np.nan, np.nan
        pos[i] = meanpos
        std[i] = stdpos
    res = np.abs(pos - zeropos) / (std * 2.3548)
    return res, pos, std


def resolvingpower_from_photonlist_robust(lphotons, orders,
                                          cols, zeropositions,
                                          ordercol='order'):
    '''Robustly calculate the resolving power for grating orders

    This function wraps `resolvingpower_from_photonlist` to iterate
    over several columns. It calculates the resolving power R for several
    columns and reports the worst case.

    We often want to determine R from the distribution of photons on a CCD.
    Geometry effects like CCDs that are slightly off the
    Rowland torus make R calculated on the detectors lower than an ideal case.
    However, that R can spike if part of the line hits a
    chip gap. So, we can calculate R for a continuous detector on the
    ideal Rowland circle and report that R in such a case.

    Parameters
    ----------
    lphotons : list of `astropy.table.Table`
        List of Photon event list
    orders : np.array
        Orders for which the resolving power will be calculated
    cols : list
        List of column names for the column holding the dispersion coordinate.
    zeropositions : list
        List of values of column `col` where the zeroth order is found. If not given,
        this is calculated (assuming the zeroth order photons are part of the
        event list).
    ordercol : string
        Name of column that lists grating order for each photon

    Returns
    -------
    res : np.array
        resolving power for each order
    pos : np.array
        mean value of ``col`` for each order
    std : np.array
        standard deviation of the distribution of ``col`` for each order
    '''
    if len(cols) != len(zeropositions):
        raise ValueError('Number of elements in cols and zeropositions is not the same.')
    if len(cols) != len(lphotons):
        raise ValueError('Number of elements in cols and photon lists is not the same.')

    results = np.zeros((len(cols), len(orders)))
    positions = np.zeros_like(results)
    stds = np.zeros_like(results)
    for i, (photons, col, zeropos) in enumerate(zip(lphotons, cols, zeropositions)):
        res, pos, std = resolvingpower_from_photonlist(photons, orders,
                                                    col=col, zeropos=zeropos,
                                                    ordercol=ordercol)
        results[i, :] = res
        positions[i, :] = pos
        stds[i, :] = std
    ind = np.argmin(results, axis=0)
    ind_n = np.arange(results.shape[1])
    return results[ind, ind_n], positions[ind, ind_n], stds[ind, ind_n]


def effectivearea_from_photonlist(photons, orders, n_photons, A_geom=1. * u.cm**2,
                                  ordercol='order'):
    '''Calculate the effective area several grating orders

    This is based on the probabilities of the photons in the list, so
    this photons list must already account for all instrument
    components, for example, do not forget to set the probability to 0
    for photons that fall into a chip gap.

    Parameters
    ----------
    photons : `astropy.table.Table`
        Photon event list
    orders : np.array
        Orders for which the resolving power will be calculated
    n_photons : int
        Number of photons originally simulated
    A_geom : quantity
        Geometric area of aperture that was used for photon list.
    ordercol : string
        Name of column that lists grating order for each photon

    Returns
    -------
    aeff : np.array
        effective area for each order
    '''
    aeff = np.zeros(len(orders))
    for i, o in enumerate(orders):
        aeff[i] = photons['probability'][photons[ordercol] == o].sum()
    return aeff / n_photons * A_geom


def identify_photon_in_subaperture(angle, max_ang, ang_0=np.pi / 2):
    '''Find which photons are in a sub-aperture of two mirrored sectors

    This function includes mirroring on the 0 line, e.g. if
    max_ang=20 deg and ang_0 = 90 deg, includes photons in two sectors
    from 70-110 deg and 250-290 deg.

    Parameters
    ----------
    angle : array
        Angles (in rad) to be tested
    max_ang : float
        Maximal distance (in rad) from ``ang_0`` that is included in the sub-aperturing
    ang_0 : float
        Center of the sub-aperturing region. Default is sub-aperturing perpendicular
        to the ``angle=0`` line, which is the most common sub-aperturing layout with
        subaperturing perpendicular to the dispersion direction as ``angle=0``.
    '''
    indplus = np.isclose(np.mod(angle, 2 * np.pi), ang_0, atol=max_ang)
    indminus = np.isclose(np.mod(angle, 2 * np.pi), 2 * np.pi - ang_0, atol=max_ang)
    return indplus | indminus


class CaptureResAeff():
    '''Capture resolving power and effective area for a simulation.

    Instances of this class can be called with a photon list for a
    simulation. The photon list will be analyzed for resolving power and
    effective area in a number of relevant orders.

    This is implemented as a class and not a simple function. When the class is
    initialized a number of parameters that are true for any to the
    analysis (e.g. the names of certain columns) are set and saved in the class
    instance.

    The implementation of this class is geared towards instruments with
    gratings but can also serve as an example how a complex analysis that
    derives several different parameters can be implemented.

    Results for effective area and resolving power are reported on a per order
    basis and also summarized for all grating orders combined and the zeroth
    order separately.

    Inherit form this class to customize the filters applied to the photon list
    before effective area and resolving power are calculated.

    Parameters
    ----------
    A_geom : number
        Geometric area of aperture for the simulations that this instance
        will analyze.
    order_col : string
        Column names for grating orders
    orders : array
        Order numbers to consider in the analysis
    dispersion_coord : string
        Dispersion coordinate for
        `marxs.analysis.gratings.resolvingpower_from_photonlist`. Any photons
        that have non-finite values or are masked in this column will be ignored
        for the purpose of the resolving power calculation.
    zeropos : float
        Position of zeroth order in ``dispersion_coord``
    '''
    def __init__(self, A_geom=1, order_col='order',
                 orders=np.arange(-10, 11),
                 dispersion_coord='det_x',
                 zeropos=None):
        self.A_geom = A_geom
        self.order_col = order_col
        self.orders = np.asanyarray(orders)
        self.dispersion_coord = dispersion_coord
        self.zeropos = zeropos

    def aeff_filter(self, photons):
        '''Filter photon list before calculating the effective area.

        Here: No-op (all photons are used.)

        Parameters
        ----------
        photons : `astropy.table.Table`
            Photon list.

        Returns
        -------
        ind : numpy array
           array of boolean values that can be used to index ``photons`` and select
           a sub-set to be used to calculate the effective area.
        '''
        return np.ones(len(photons), dtype=bool)

    def res_filter(self, photons):
        ind = (np.isfinite(photons[self.dispersion_coord]) &
                (photons['probability'] > 0))
        if hasattr(photons[self.dispersion_coord], 'mask'):
            ind = ind & ~photons[self.dispersion_coord].mask
        return ind

    def __call__(self, photons, n_photons=None):
        '''Calculate Aeff and R for an input photon list.

        Parameters
        ----------
        photons : `astropy.table.Table`
            Photon list.
        n_photons : int or `None`
            Number of photons originally simulated. If ``None`` use length of
            ``photons``.

        Returns
        -------
        result : dict
            Dictionary with per-order Aeff and R, as well as values
            summed over all grating orders.
        '''
        if n_photons is None:
            n_photons = len(photons)
        ind = self.aeff_filter(photons)
        aeff = effectivearea_from_photonlist(photons[ind], self.orders, n_photons,
                                             self.A_geom, self.order_col)
        try:
            ind = self.res_filter(photons)
            res, pos, std = resolvingpower_from_photonlist(photons[ind], self.orders,
                                                           col=self.dispersion_coord,
                                                           zeropos=self.zeropos,
                                                           ordercol=self.order_col)
        except AnalysisError:
            # Something did not work, e.g. too few photons to find zeroth order
            res = np.nan * np.ones(len(self.orders))

        disporders = self.orders != 0
        avggratres, aeffgrat = average_R_Aeff(res[disporders],
                                              aeff[disporders])
        aeff0 = np.sum(aeff[~disporders])
        return {'Aeff0': aeff0, 'Aeffgrat': aeffgrat, 'Aeff': aeff,
                'Rgrat': avggratres, 'R': res}


class CaptureResAeff_CCDgaps(CaptureResAeff):
    '''Capture resolving power and effective area for a tolerancing simulation.

    Unlike `CaptureResAeff` this objects is set up to take into account CCD gaps.
    For the resolving power calculation, one wants to ignore CCD gaps, because they
    might make the observed LSF artificially small - if only half of the LSF falls
    on a CCD, and the other photons are not counted, the LSF will appear only
    half as wide. In contrast, for the effective area calculation, we *do* want to
    ignore photons that miss a CCD.

    Here, we following approach is taken: This class overrides
    `~marxs.design.tolerancing.CaptureResAeff.aeff_filter`.
    All photons with a probability >0 are presumed ot reach the focal plane.
    For the effective area, an additional
    filter is applied: Only photons with `photons['aeff_filter_col'] > 0` will
    be used. Typically, this could be the CCD_ID.

    Parameters
    ----------
    aeff_filter_col : string
        Only photons where the value in this column is larger than 0 will be used
        for the calculation of the effective area.

    kwargs :
        All other keyword arguments are passed to `CaptureResAeff`.
    '''
    def __init__(self, aeff_filter_col='CCD_ID', **kwargs):
        super().__init__(**kwargs)
        self.aeff_filter_col = aeff_filter_col

    def aeff_filter(self, photons):
        '''Filter photon list before calculating the effective area.

        Only photons with ``photons[aeff_filter_col] >= 0`` are used.

        Parameters
        ----------
        photons : `astropy.table.Table`
            Photon list.

        Returns
        -------
        ind : numpy array
           array of boolean values that can be used to index ``photons`` and select
           a sub-set to be used to calculate the effective area.
        '''
        return photons[self.aeff_filter_col] >= 0


# Licensed under GPL version 3 - see LICENSE.rst
from tempfile import TemporaryDirectory
from glob import glob
from os.path import join as pjoin
import numpy as np
import astropy.units as u
from astropy.table import Table, QTable

from marxs.missions.arcus import arfrmf
from marxs.missions.arcus.utils import config as arcusconf
from marxs.optics.detector import CCDRedistNormal
from ..osip import FixedWidthOSIP, FixedFractionOSIP, FractionalDistanceOSIP

import pytest


# In this example, we use a file that's included in MARXS for test purposes
from astropy.utils.data import get_pkg_data_filename
filename = get_pkg_data_filename('data/ccd_redist_normal.ecsv', 'marxs.optics.tests')
tab_redist = QTable.read(filename)

ccd_redist = CCDRedistNormal(tab_width=tab_redist)
osipf = FixedWidthOSIP(40 * u.eV, ccd_redist=ccd_redist)
osipp = FixedFractionOSIP(0.7, ccd_redist=ccd_redist)
osipd = FractionalDistanceOSIP(ccd_redist=ccd_redist)


@pytest.mark.parametrize("thisosip", [osipf, osipp, osipd])
def test_osip_tab_returns(thisosip):
    '''Check return format and unit'''
    wave = [20, 21, 22, 23] * u.Angstrom
    tab = thisosip.osip_tab(wave, 2)

    assert tab.shape == (2, 4)
    assert tab.unit.physical_type == 'energy'
    assert np.all(tab.value > 0)


def test_osip_factor_width_unit():
    '''check units work for osip_width'''
    osip1 = FixedWidthOSIP(40 * u.eV, ccd_redist=ccd_redist)
    osip2 = FixedWidthOSIP(0.04 * u.keV, ccd_redist=ccd_redist)
    assert np.all(osip1.osip_factor([10] * u.Angstrom, -5, -5) ==
                  osip2.osip_factor([10] * u.Angstrom, -5, -5))


@pytest.mark.parametrize("thisosip", [osipf, osipp, osipd])
def test_osip_factor_wave_unit(thisosip):
    '''Check input works for wave or energy'''
    wave = [20, 21, 22, 23] * u.Angstrom
    energ = wave.to(u.keV, equivalencies=u.spectral())
    assert np.all(thisosip.osip_factor(wave, -5, -5) ==
                  thisosip.osip_factor(energ, -5, -5))


def test_osip_factor():
    '''test extreme values that do not depend on CCD resolution'''
    assert FixedWidthOSIP(0 * u.eV, ccd_redist=ccd_redist).osip_factor([10] * u.Angstrom, -5, -5) == 0
    wide = FixedWidthOSIP(10 * u.keV, ccd_redist=ccd_redist)
    assert np.allclose(wide.osip_factor([10] * u.Angstrom, -5, -5), 1)
    '''test with fixed sigma'''
    sig = QTable({'energy': [0, 1, 2] * u.keV, 'sigma': [40, 40, 40] * u.eV})
    myosip = FixedWidthOSIP(40 * u.eV, ccd_redist=CCDRedistNormal(tab_width=sig))

    assert np.allclose(myosip.osip_factor([10] * u.Angstrom, -5, -5),
                       0.6827, rtol=1e-4)


@pytest.mark.parametrize("thisosip", [osipf, osipp, osipd])
def test_osip_factor_orders_on_different_sides(thisosip):
    '''If one order is positive and the other negative, then
    the signal is diffracted to opposite sides, so there is no
    contamination.
    '''
    assert thisosip.osip_factor([10] * u.Angstrom, -5, 5) == 0
    assert thisosip.osip_factor([10] * u.Angstrom, 1, -1) == 0
    assert thisosip.osip_factor([10] * u.Angstrom, 1, 0) == 0


@pytest.mark.parametrize("thisosip", [osipf, osipp, osipd])
def test_symmetry(thisosip):
    '''offset_orders +1 and -1 should have the same OSIP factors
    if OSIP is symmetric'''
    up = thisosip.osip_factor([10, 20, 30] * u.Angstrom, -5, -6)
    down = thisosip.osip_factor([10, 20, 30] * u.Angstrom, -5, -4)
    assert np.allclose(up, down, atol=1e-4)


@pytest.mark.parametrize("frac", [.5, .7, 1.])
def test_FractionalDistance_boundaries(frac):
    '''Test the boundaries of the extraction regions'''
    osipd = FractionalDistanceOSIP(frac, ccd_redist=ccd_redist)
    mlamgrid = [120, 130, 140] * u.Angstrom


    # Make comparison at the same m lambda position, i.e. the same poisition
    # on the chip
    tab5 = osipd.osip_tab(mlamgrid / 5, -5).to(u.eV, equivalencies=u.spectral())
    tab6 = osipd.osip_tab(mlamgrid / 6, -6).to(u.eV, equivalencies=u.spectral())

    # At the same mlambda, the dE does not depend on order
    assert u.allclose(tab5, tab6)

    # Now actually get the dE
    mlam5en = (mlamgrid / 5).to(u.eV, equivalencies=u.spectral())
    mlam6en = (mlamgrid / 6).to(u.eV, equivalencies=u.spectral())

    # Areas covered by extraction region
    a5 = tab5[0, :]
    a6 = tab6[1, :]

    assert np.allclose((a5 + a6) / (mlam6en - mlam5en), frac)


def test_Fixed_width_width():
    tab = osipf.osip_tab([10, 20, 30] * u.Angstrom, 5)
    assert u.allclose(tab, 40 * u.eV)


@pytest.mark.skipif('data' not in arcusconf,
                    reason='Test requires Arcus CALDB')
def test_Fixed_Fraction():
    '''Test FixedFractionOSIP end-to-end by actually making and
    changing a file'''

    with TemporaryDirectory() as tmpdirname:
        # -5 is last, because we need that for the tests below
        for o in [-4, -6, -5]:
            arf = arfrmf.mkarf([23, 24, 25, 26] * u.Angstrom, o)
            arf = arfrmf.tagversion(arf)

            basearf = pjoin(tmpdirname,
                            arfrmf.filename_from_meta('arf', **arf.meta))
            arf.write(basearf)

        osipp = FixedFractionOSIP(0.8, ccd_redist=ccd_redist,
                                  filename_from_meta=arfrmf.filename_from_meta)
        osipp.apply_osip_all(tmpdirname, tmpdirname, [-5],
                             filename_from_meta_kwargs={'ARCCHAN': '1111'})

        arfcenter = Table.read(glob(pjoin(tmpdirname, '*-5*-5*'))[0],
                               format='fits')
        assert np.isclose(arfcenter.meta['OSIPFAC'], 0.8)
        arfup = Table.read(glob(pjoin(tmpdirname, '*-5*-4*'))[0],
                           format='fits')
        assert np.all(arfup['OSIPFAC'] < .1)

        # Narrower OSIP region gives lower effective area
        osipp = FixedFractionOSIP(0.7, ccd_redist=ccd_redist,
                                  filename_from_meta=arfrmf.filename_from_meta)
        osipp.apply_osip_all(tmpdirname, tmpdirname, [-5], outroot='root',
                             filename_from_meta_kwargs={'ARCCHAN': '1111'})

        # To help debug the test:
        allfiles = glob(pjoin(tmpdirname, '*'))

        arfcenter7 = Table.read(glob(pjoin(tmpdirname, 'root*-5*-5*'))[0],
                                format='fits')
        assert np.isclose(arfcenter7.meta['OSIPFAC'], 0.7)
        assert u.allclose(arfcenter7['SPECRESP'],
                          0.7 / .8 * arfcenter['SPECRESP'])
        arfup7 = Table.read(glob(pjoin(tmpdirname, 'root*-5*-4*'))[0],
                            format='fits')
        assert np.all(arfup7['OSIPFAC'] < arfup['OSIPFAC'])



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.table import Table, QTable
import astropy.units as u
import transforms3d
import pytest

from ..detector import FlatDetector, CircularDetector, CCDRedistNormal
from ...tests import closeornan
from ...math.utils import h2e
from ...design import RowlandTorus
from ...math.geometry import Cylinder
from ...utils import generate_test_photons


def test_pixelnumbers():
    pos = np.array([[0, 0., -0.25, 1.],
                    [0., 9.9, 1., 1.],
                    [0., 10.1, 1., 1.]])
    dir = np.ones((3,4), dtype=float)
    dir[:, 3] = 0.
    photons = Table({'pos': pos, 'dir': dir,
                     'energy': [1,2., 3.], 'polarization': [1.,2, 3.], 'probability': [1., 1., 1.]})
    det = FlatDetector(zoom=[1., 10., 5.], pixsize=0.5)
    assert det.npix == [40, 20]
    assert det.centerpix == [19.5, 9.5]
    photons = det(photons)
    assert closeornan(photons['det_x'], np.array([0, 9.9, np.nan]))
    assert closeornan(photons['det_y'], np.array([-0.25, 1., np.nan]))
    assert closeornan(photons['detpix_x'], np.array([19.5, 39.3, np.nan]))
    assert closeornan(photons['detpix_y'], np.array([9, 11.5, np.nan]))

    # Regression test: This was rounded down to [39, 39] at some point.
    det = FlatDetector(pixsize=0.05)
    assert det.npix == [40, 40]


def test_nonintegerwarning(recwarn):
    det = FlatDetector(zoom=np.array([1.,2.,3.]), pixsize=0.3)
    w = recwarn.pop()
    assert 'is not an integer multiple' in str(w.message)


def test_CircularDetector_from_Rowland():
    '''If a circular detector is very narrow, then all points should be
    very close to the Rowland torus.'''
    rowland = RowlandTorus(R=6e4, r=5e4, position=[123., 345., -678.],
                           orientation=transforms3d.euler.euler2mat(1, 2, 3, 'syxz'))
    detcirc = CircularDetector(geometry=Cylinder.from_rowland(rowland, width=1e-6))
    phi = np.mgrid[0:2.*np.pi:360j]
    points = detcirc.geometry.parametric_surface(phi)
    # Quartic < 1e5 is very close for these large values of r and R.
    assert np.max(np.abs(rowland.quartic(h2e(points)))) < 1e5


def test_CCD_Redist_Normal():
    """We're not checking the implementation of a normal function, but just that the
    right columns are read and used."""
    photons = generate_test_photons(100)
    photons["energy"][:50] = 1
    photons["energy"][50:] = 5

    res = QTable({"energy": [1, 10] * u.keV, "sigma": [0.1, 1.0] * u.keV})

    ccdre = CCDRedistNormal(tab_width=res)
    photons = ccdre(photons)

    assert "energy_detected" in photons.colnames
    assert np.mean(photons["energy_detected"][:50]) == pytest.approx(1.0, rel=0.02)
    assert np.mean(photons["energy_detected"][50:]) == pytest.approx(5.0, rel=0.1)
    assert np.std(photons["energy_detected"][:50]) == pytest.approx(0.1, rel=0.2)
    assert np.std(photons["energy_detected"][50:]) == pytest.approx(0.5, rel=0.2)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np

from marxs.source import PointSource, FixedPointing
import astropy.units as u
from astropy.coordinates import SkyCoord
from marxs.math.utils import xyz2zxy
from .. import Arcus

import pytest

e = 0.5 * u.keV

mysource = PointSource(coords=SkyCoord(30. * u.deg, 30. * u.deg),
                       energy=e)
mypointing = FixedPointing(coords=SkyCoord(30 * u.deg, 30. * u.deg),
                           reference_transform=xyz2zxy)


@pytest.mark.parametrize("instrument", [Arcus(channels=['1']),
                                        Arcus(channels=['2m'])])
def test_orders_are_focussed(instrument):
    '''Check that the orders look reasonable.

    This test tries to be generic so that coordinate system etc. can be
    changed later, but still check that all light is focused to
    one point to detect error in setting up the mirrors.
    '''
    photons = mysource.generate_photons(2e4 * u.s)
    photons = mypointing(photons)
    photons = instrument(photons)

    for i in range(-12, 1):
        ind = (photons['order'] == i) & np.isfinite(photons['det_x']) & (photons['probability'] > 0)
        if ind.sum() > 100:
            assert np.std(photons['det_y'][ind & (photons['order_L1'] == 0)]) < 1
            assert np.std(photons['det_x'][ind]) < 1
            assert np.std(photons['det_x'][ind]) < np.std(photons['det_y'][ind])
            # But make sure that are not focussed too much
            # That would indicate that the scattering did not work
            assert np.std(photons['det_x'][ind]) > .04
            assert np.std(photons['det_y'][ind]) > .1


def test_zeroth_order_and_some_dispersed_orders_are_seen():
    '''test that both the zeroth order and some of the dispersed
    orders are positioned in the detector.
    '''
    photons = mysource.generate_photons(2e4 * u.s)
    photons = mypointing(photons)
    photons = Arcus()(photons)

    n_det = [((photons['order'] == i) & np.isfinite(photons['det_x'])).sum() for i in range(-12, 1)]
    assert n_det[-1] > 0
    assert sum(n_det[:9]) > 0


def test_two_optical_axes():
    '''Check that there are two position for the zeroth order.'''
    photons = mysource.generate_photons(1e5 * u.s)
    photons = mypointing(photons)
    photons = Arcus()(photons)
    i0 = (photons['order'] == 0) & np.isfinite(photons['det_x']) & (photons['probability'] > 0)

    assert i0.sum() > 250
    assert np.std(photons['det_y'][i0]) > 1
    assert np.std(photons['det_x'][i0]) > 1


@pytest.mark.parametrize("instrum, expected_area",
                         [(Arcus(), 400 * u.cm**2),
                          (Arcus(channels=['1']), 100 * u.cm**2),
                          (Arcus(channels=['2']), 100 * u.cm**2),
                          (Arcus(channels=['1m']), 100 * u.cm**2),
                          (Arcus(channels=['2m']), 100 * u.cm**2)])
def test_effective_area(instrum, expected_area):
    '''Surely, the effective area of Arcus will evolve a little when the
    code is changed to accommodate e.g. a slightly different mounting
    for the gratings, but if the effective area drops or increases
    dramatically, that is more likely a sign for a bug in the code.
    '''
    photons = mysource.generate_photons(2e4 * u.s)
    photons = mypointing(photons)
    photons = instrum(photons)
    ind = np.isfinite(photons['det_x'])
    a_eff = np.sum(photons['probability'][ind]) / len(photons) * instrum.elements[0].area
    assert a_eff > 0.7 * expected_area
    assert a_eff < 1.5 * expected_area



from astropy.table import Table
import astropy.units as u


class DataFileFormatException(Exception):
    pass


def load_number(filename, valuename):
    '''Get a single number from an ecsv input file

    Parameters
    ----------
    filename : string
        Name of data file
    valuename : string
        Name of the column that hold the value

    Returns
    -------
    val : float or `astropy.units.Quantity`
        If the unit of the column is set, returns a `astropy.units.Quantity`
        instance, otherwise a plain float.
    '''
    tab = Table.read(filename, format='ascii.ecsv')
    if len(tab) != 1:
        raise DataFileFormatException('Table {} contains more than one row of data.'.format(filename))
    else:
        if tab[valuename].unit is None:
            return tab[valuename][0]
        else:
            return u.Quantity(tab[valuename])[0]


def load_table(filename):
    '''Get a table from an ecsv input file

    Parameters
    ----------
    filename : string
        Name of data file

    Returns
    -------
    val : `astropy.table.Table`
    '''
    tab = Table.read(filename, format='ascii.ecsv')
    return (tab)




# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
import astropy.units as u
from scipy.interpolate import RectBivariateSpline

from ... import source, optics
from ...math.utils import h2e
from marxs.utils import generate_test_photons


@pytest.mark.parametrize("ra", [0., 30., -60.])
def test_PerfectLens(ra):
    mysource = source.PointSource(coords=SkyCoord(0., ra, unit="deg"))
    mypointing = source.FixedPointing(coords=SkyCoord(0., 0., unit='deg'))
    myslit = optics.RectangleAperture(zoom=2, position=[+100, 0, 0])
    f = 1000
    lens = optics.PerfectLens(focallength=f, zoom=400)

    photons = mysource.generate_photons(11 * u.s)
    photons = mypointing(photons)
    photons = myslit(photons)
    assert np.allclose(h2e(photons['pos'].data)[:, 0], 100.)

    photons = lens(photons)
    assert np.allclose(h2e(photons['pos'].data)[:, 0], 0.)

    # How far away do I need to put a detector to hit the focal point?
    d = f * np.cos(np.deg2rad(ra))
    mdet = optics.FlatDetector(pixsize=0.01, position=np.array([-d, 0, 0]), zoom=1e5)
    photons = mdet(photons)
    assert np.std(photons['det_x']) < 1e-4
    assert np.std(photons['det_y']) < 1e-4

    # Check that they actually diverge for other detector placement
    mdet = optics.FlatDetector(pixsize=0.01, position=np.array([-2 * d, 0, 0]), zoom=1e5)
    photons = mdet(photons)
    assert np.std(photons['det_x']) > 1e-4
    assert np.std(photons['det_y']) > 1e-4


def test_PerfectLens_offset():
    """Test a lens that is not centered on the optical axis."""
    lens = optics.PerfectLens(focallength=100, position=[0, 0, 10], zoom=400)
    lens2 = optics.PerfectLens(
        focallength=100, position=[0, 0, 0], zoom=400, d_center_optical_axis=-10
    )
    lens3 = optics.PerfectLens(
        focallength=100, position=[0, 0, 0], zoom=400, d_center_optical_axis=0
    )
    photons = generate_test_photons(1)
    p2 = generate_test_photons(1)
    p3 = generate_test_photons(1)
    photons = lens(photons)
    p2 = lens2(p2)
    p3 = lens3(p3)
    np.testing.assert_allclose(photons["dir"], p2["dir"])
    not np.allclose(photons["dir"], p3["dir"])


def test_PerfectLens_refl():
    """Test a reflection.

    This test sis a bit silly in the sense that it
    uses an unnecessarily complicated reflectivity function
    but then reflect at 0 deg."""
    xarr = np.linspace(-3, 3, 100)
    yarr = np.linspace(-3, 3, 100)
    xgrid, ygrid = np.meshgrid(xarr, yarr, indexing="ij")
    zdata = np.exp(-np.sqrt((xgrid / 2) ** 2 + ygrid**2))
    interp = RectBivariateSpline(xarr, yarr, zdata, kx=1, ky=1)

    lens = optics.PerfectLens(focallength=123.0, reflectivity_interpolator=interp)
    photons = generate_test_photons(1)
    photons = lens(photons)

    assert np.allclose(photons["energy"], 1)
    assert np.allclose(photons["polarization"], [0, 1, 0, 0])
    assert np.allclose(photons["probability"], 0.367205)
